#' DGP for simulation design
#' @param sigma \eqn{sd(Y|X)}
#' @param n number of observations
#' @param M bound on second derivative of f(X)
#' @param b knots for spline. Has length 3 and ranges [0, 1] for symmetric
#'     design, has length 4 and ranges [-1, 1] for asymmetric design
#' @keywords internal
LPDGP <- function(sigma, n, M, b) {
    X <- 2 * stats::runif(n) - 1    # uniform[-1, 1]

    ## symmetric spline
    sf <- function(X, b)
        X^2 - 2*pmax(abs(X)-b[1], 0)^2 + 2*pmax(abs(X)-b[2], 0)^2 - 2 *
            pmax(abs(X)-b[3], 0)^2

    #' Asymmetric spline
    sfa <- function(X, b)
        (X+1)^2 - 2*pmax(X-b[1], 0)^2 + 2*pmax(X-b[2], 0)^2 -
            2*pmax(X-b[3], 0)^2 + 2*pmax(X-b[4], 0)^2
    f <- if (length(b)==4) sfa(X, b)-sfa(0, b) else sf(X, b)

    data.frame(y=M/2 * f + sigma*stats::rnorm(n), x=X)
}

#' Compute confidence intervals and estimates using a variety of methods
#' @param Mh vector of values of M for Holder-honest CIs
#' @keywords internal
LPGetStats <- function(d, Mh) {
    ## CCF at CCF defaults
    if (!requireNamespace("nprobust", quietly = TRUE)) {
        stop("Please install nprobust package;
              it's needed performing Monte Carlo.",
             call. = FALSE)
    }
    d <- RDHonest::LPPData(d, point=0)
    bw.rot <- RDHonest::ROTBW.fit(d, kern="triangular", order=1, boundary=FALSE)
    ccf.rot <- nprobust::lprobust(y=d$Y, x=d$X, eval=0, kernel="tri",
                                  interior=TRUE, h=bw.rot, vce="nn")
    ccf.mse <- nprobust::lprobust(y=d$Y, x=d$X, eval=0, kernel="tri",
                                  interior=TRUE, bwselect="mse-dpi", vce="nn")
    ccf.ce <- nprobust::lprobust(y=d$Y, x=d$X, eval=0, kernel="tri",
                                 interior=TRUE, bwselect="ce-dpi", vce="nn")

    r <- data.frame(method=c(rep(c("Naive", "RBC"), 3)),
                    bw=c(rep("MSE", 2), rep("CE", 2), rep("ROT", 2)),
                    bias=c(ccf.mse$Estimate[1, "tau.us"],
                           ccf.mse$Estimate[1, "tau.bc"],
                           ccf.ce$Estimate[1, "tau.us"],
                           ccf.ce$Estimate[1, "tau.bc"],
                           ccf.rot$Estimate[1, "tau.us"],
                           ccf.rot$Estimate[1, "tau.bc"]),
                    se=c(ccf.mse$Estimate[1, "se.us"],
                         ccf.mse$Estimate[1, "se.rb"],
                         ccf.ce$Estimate[1, "se.us"],
                         ccf.ce$Estimate[1, "se.rb"],
                         ccf.rot$Estimate[1, "se.us"],
                         ccf.rot$Estimate[1, "se.rb"]),
                    h=c(rep(ccf.mse$Estimate[1, "h"], 2),
                        rep(ccf.ce$Estimate[1, "h"], 2),
                        rep(ccf.rot$Estimate[1, "h"], 2)))
    za <- qnorm(0.975)
    r$hl <- r$se[1:6]*za
    r$r <- abs(r$bias/r$se) > za

    ## Honest inference:
    d1 <- RDHonest::LPPPrelimVar(d, se.initial="ROTEHW")
    if (!is.null(Mh)) {
        MROT <- RDHonest::LPP_MROT.fit(d)
        Ms <- c(Mh, MROT, 2*MROT)

        for (j in seq_along(Ms)) {
            h <- RDHonest::LPPHonest.fit(d1, M=Ms[j], sclass="H", J=3,
                                         opt.criterion="MSE", se.method="nn")
            Mlab <- if (j >length(Mh)) paste0("ROT", j-length(Mh)) else Ms[j]
            r <- rbind(r, data.frame(method=paste0("Hon, C=", Mlab),
                                     bw="MSE", bias=h$estimate,
                                     se=h$sd, h=h$h, hl=h$hl,
                                     r=abs(h$estimate)>h$hl))
        }
    }
    rownames(r) <- NULL
    r
}

#' Monte Carlo simulation of single design for coverage and length of CIs.
#'
#' Simulate CI coverage and length, comparing pointwise and uniform methods.
#' @param Mh vector of values of M for Holder-honest CIs
#' @param nsim number of simulation draws
#' @inheritParams RD_DGP
#' @keywords internal
LPSimulateSingle <- function(sigma, n, M, b, nsim, cores=NULL, Mh) {
    ## get dims of output
    se <- LPGetStats(LPDGP(sigma, n, M, b), Mh)
    r <- array(dim=c(nrow(se), ncol(se)-2, nsim))
    set.seed(42)
    d <- LPDGP(sigma, n*nsim, M, b)
    Y <- matrix(d$y, nrow=n, ncol=nsim)
    X <- matrix(d$x, nrow=n, ncol=nsim)

    if (is.null(cores))
        cores <- parallel::detectCores()-1L     # number of cores

    SimDraw <- function(j)
        as.matrix(LPGetStats(data.frame(y=Y[, j], x=X[, j]), Mh)[, -c(1, 2)])

    r <- if (cores>1) {
             parallel::mclapply(seq_len(nsim), SimDraw, mc.cores=cores)
         } else {
             lapply(seq_len(nsim), SimDraw)
         }
    r <- array(unlist(r), dim=c(nrow(r[[1]]), ncol(r[[1]]), length(r)))

    ra <- cbind(se[, 1:2], rowMeans(r, dims=2))
    names(ra) <- names(se)
    ra
}


#' Monte Carlo simulation for coverage and length of CIs for a set of designs
#'
#' Simulate CI coverage and length, comparing pointwise and uniform methods
#' using a number of designs
#' @param p Data frame with design parameters.
#' @param Mh vector of values of M for Holder-honest CIs
#' @param nsim number of simulation draws
#' @param cores number of CPU cores (for parallel computation). Put
#'     \code{cores=0} for serial computation only.
#' @param eff.row Half-legnth comparisons should be wrt which row?
#' @inheritParams RD_DGP
#' @export
LPSimulate <- function(sigma=NULL, n=NULL, M, p, nsim, eff.row=1,
                       cores=NULL, Mh=NULL) {

    s <- Sys.time()
    if(is.null(cores))
        cores <- parallel::detectCores()-1L     # number of cores
    if(cores>1 & nrow(p)<20) {
        ## parallelize monte carlo, not searching through designs
        scores <- cores
        cores <- 1
    } else {
        scores <- 0
    }

    single <- function(j) {
        rj <- LPSimulateSingle(sigma=sigma, n=n, M=M, b=p[j, ],
                               nsim=nsim, cores=scores, Mh)
        rj$eff <- rj$hl/rj$hl[eff.row]
        cbind(as.data.frame(lapply(p[j, ], function(a)
            rep(a, nrow(rj)))), rj)
    }

    r <- if (cores>1) {
             parallel::mclapply(seq_len(nrow(p)), single, mc.cores=cores)
         } else {
             lapply(seq_len(nrow(p)), single)
         }

    print(difftime(Sys.time(), s))

    do.call(rbind, r)
}

#' Plot functions used in Monte Carlo design
#'
#' @param M bound on second derivative of f(X)
#' @param b list of b specifications
#' @param legends optional vector of legend labels
#' @export
LPPlotDesign <- function(M, b, legends=NULL) {
    d <- lapply(1:length(b), function(j) {
        r <- LPDGP(sigma=0, n=1000, M, b[[j]])
        r[order(r$x), ]
    })
    d <- do.call(rbind, d)
    if (is.null(legends)) {
        legends <- sapply(1:length(b), function(j)
            paste0("$b_{1}=", round(b[[j]][1], 2), "$, $b_{2}=",
                   round(b[[j]][2], 2),
                   "$", ", $b_{3}=", round(b[[j]][3], 2),
                   if (length(b[[j]])==4)
                       paste0("$, $b_{4}=", round(b[[j]][4], 2), "$") else "$"))
    }
    d <- data.frame(x=d$x, y=d$y, what=rep(legends, each=nrow(d)/length(b)))

    baseplot(d, ylab="$f(x)$", xlab="$x$",
             method=list(cex=0.9, "first.bumpup"))
}
