## Graph minimax estimators and confidence intervals in RD problem with Taylor
## smoothness class, and estimators based on local linear regression under
## Taylor or Holder class

#' Construct table of minimax estimates under Taylor(2) class
#'
#' Construct table of minimax estimates and CIs under Taylor class for given
#' values of smoothness parameter \eqn{C}
#' @param d RDData
#' @param Cs vector of smoothness parameters
#' @keywords internal
tbl_RDTOpt <- function(d, Cs, se.method) {

    tbl <- function(opt.criterion="OCI", bound=FALSE, alpha=0.05, beta=0.5) {
        df <- data.frame()
        for (C in Cs) {
            r <- if (bound==TRUE) {
                RDHonest::RDTEfficiencyBound(d, 2*C, opt.criterion,
                                             alpha, beta)
            } else {
                RDHonest::RDTOpt.fit(d, 2*C, opt.criterion, alpha, beta,
                                     se.method)
            }
            df <- rbind(df, data.frame(r[-2], C=C)) # drop lff
        }
        df
    }

    list(oci   = tbl(opt.criterion="OCI", beta=0.8),
         e     = tbl(opt.criterion="MSE"),
         flci  = tbl(opt.criterion="FLCI"),
         eoci5 = tbl(opt.criterion="OCI", bound=TRUE),
         eoci8 = tbl(opt.criterion="OCI", bound=TRUE, beta=0.8),
         eflci = tbl(opt.criterion="FLCI", bound=TRUE))
}

#' Construct table of optimal estimates based on local polynomial regression
#'
#' Construct table of optimal estimates and CIs based on local linear regression
#' under Taylor or Holder class for given values of smoothness parameter \eqn{C}
#' @param Cs vector of smoothness parameters
#' @param d RDData
#' @inheritParams RDHonest::RDHonest.fit
#' @keywords internal
tbl_RDLP <- function(d, Cs, kern, sclass="T", bw.equal=FALSE,
                     order=1, se.method) {
    rdh <- function(C, what)
        RDHonest::RDHonest.fit(d, 2*C, kern, opt.criterion=what, alpha=0.05,
                               beta=0.8, bw.equal=bw.equal, order=order,
                               se.method=se.method, sclass=sclass)

    tbl <- function(what="OCI") {
        df <- data.frame()
        for (C in Cs) {
            df <- rbind(df, data.frame(rdh(C, what)[-2], C=C)) # drop lff
        }
        df
    }

    list(oci  = tbl(what="OCI"),
         e    = tbl(what="MSE"),
         flci = tbl(what="FLCI"))
}

#' replaces plotEstimates and InferencePlot
#' @param d list of data frames as returned by tbl_RDTOpt or tbl_RDLP
#' @param xticks indices of rows of \code{d$e} at which to tick the x-axis
#' @keywords internal
plot_estimates <- function(d, ylab="", xticks=NULL, xlim=NULL, labels=TRUE) {

    ## make data frame for plotting
    var <- c("one-sided", "one-sided ", "two-sided", "two-sided ", "estimate")
    col <- c("one-sided", "one-sided", "two-sided", "two-sided", "estimate")
    df <- data.frame(x=rep(d$oci$C, 5),
                     y=c(d$oci$lower, d$oci$upper, d$flci$estimate-d$flci$hl,
                         d$flci$estimate+d$flci$hl, d$e$estimate),
                     var=rep(var, each=length(d$oci$C)),
                     label=rep(col, each=length(d$oci$C)))

    p <- ggplot2::qplot(x=x, y=y, data=df, geom="line", group=var,
                        color=label, linetype=label)
    p <- p + ggplot2::theme_classic() + ggplot2::ylab(ylab) +
        ggplot2::theme(legend.position = "none",
                       axis.line.x=ggplot2::element_line(color="black",
                           size=0.2, linetype="solid"),
                       axis.line.y=ggplot2::element_line(color="black",
                           size=0.2, linetype="solid")) +
        directlabels::geom_dl(ggplot2::aes(label=var), data=df,
                              method="last.bumpup")
    ## the axis.line.x and axis.line.y are needed as theme.classic has a bug in
    ## 2.1.0 version of ggplot2, see
    ## https://github.com/hadley/ggplot2/issues/1565
    ## Workaround with geom_dl, old way of doing it didn't work

    if (is.null(xlim))
        xlim <- c(min(df$x), max(df$x)+diff(range(df$x))/5)
    p <- p + ggplot2::coord_cartesian(xlim = xlim)

    if (!is.null(xticks)) {
        l <- formatC(d$e$C[xticks], format="f", digits=6, drop0trailing=TRUE)
        hp <- format(d$e$hp[xticks], digits=2, trim=TRUE)
        hm <- format(d$e$hm[xticks], digits=2, trim=TRUE)
        lb <- paste(l, paste0("(", hp, ",", hm, ")"), sep="\n")
        if (labels) {
            p <- p + ggplot2::xlab("$C$ / ($h_+, h_-)$") +
                ggplot2::scale_x_continuous(breaks=d$e$C[xticks], labels=lb)
        } else {
            p <- p + ggplot2::xlab("") +
                ggplot2::scale_x_continuous(breaks=d$e$C[xticks],
                                                 labels=NULL)
            }
    }

    p
}

#' Construct tables and graphs
#'
#' Constructs tables and graphs based on output of tbl_RDTOpt, and tbl_RDLP
#' @param d RDData
#' @param lp,opt tbl_RDTOpt, and tbl_RDLP
#' @keywords internal
RDEmpiricalAnalysis <- function(d, lp, opt=NULL, lp.type="local linear",
                                opt.type="SY") {

    if (!is.null(opt) & ("eflci" %in% names(opt))) {
        cat("\nMinimax efficiency bounds under Taylor class:\n",
            "\nMinimax median OCI: ", summary(opt$eoci5[[1]]),
            sprintf(" Min at C=%f",
                    opt$eoci5$C[which.min(opt$eoci5[[1]])]),
            "\nMinimax 80pc quantile OCI: ", summary(opt$eoci8[[1]]),
            sprintf(" Min at C=%f",
                    opt$eoci8$C[which.min(opt$eoci8[[1]])]),
            "\nFLCI relative to Pratt: ", summary(opt$eflci[[1]]),
            sprintf(" Min at C=%f",
                    opt$eflci$C[which.min(opt$eflci[[1]])]))

        cat("\n\nEfficiency of MSE-based SY CIs relative to optimal:\n\n")
        ELratio <- opt$oci$omega /
            (opt$e$sd*(qnorm(0.8)+qnorm(0.95)-opt$e$delta)+opt$e$omega)
        cat("Excess length of OCI: ",
            summary(ELratio),
            sprintf("Min at C=%f",
                    opt$e$C[which.min(ELratio)]),
            "\nEfficiency of MSE-based FLCI: ",
            summary(opt$flci$hl/opt$e$hl),
            sprintf("Min at C=%f\n",
                    opt$e$C[which.min(opt$flci$hl/opt$e$hl)]))
    }

    if (!is.null(opt)) {
        MSEratio <- (opt$e$sd^2+opt$e$maxbias^2) /
            (lp$e$sd^2+lp$e$maxbias^2)
        ## ELratio <- opt$oci$omega /
        ##     (2*lp$oci$maxbias+lp$oci$sd*(qnorm(0.95)+qnorm(0.8)))
        ELratio <- (2*opt$oci$maxbias+opt$oci$sd*(qnorm(0.95)+qnorm(0.8))) /
            (2*lp$oci$maxbias+lp$oci$sd*(qnorm(0.95)+qnorm(0.8)))
        cat("\nEfficiency of ", lp.type, "relative to", opt.type,
            "\n\nFLCI: ",
            summary(opt$flci$hl/lp$flci$hl),
            sprintf("min at C=%f",
                    opt$flci$C[which.min(opt$flci$hl/lp$flci$hl)]),
            "\nRMSE: ",
            summary(sqrt(MSEratio)),
            sprintf("min at C=%f",
                    opt$flci$C[which.min(MSEratio)]),
            "\nOCI at 80pc quantile: ",
            summary(ELratio),
            sprintf("min at C=%f\n",
                    opt$flci$C[which.min(ELratio)]))
    }

    cat("\nEfficiency of MSE-based ", lp.type, " CIs relative to optimal",
       lp.type, "CIs\n")

    ELratio <- (2*lp$oci$maxbias+lp$oci$sd*(qnorm(0.95)+qnorm(0.8))) /
        (2*lp$e$maxbias+lp$e$sd*(qnorm(0.95)+qnorm(0.8)))

    cat("\nExcess length of OCI: ",
        summary(ELratio),
        sprintf("min at C=%f",
                lp$e$C[which.min(ELratio)]),
        "\nEfficiency of MSE-based FLCI: ",
        summary(lp$flci$hl/lp$e$hl),
        sprintf("min at C=%f\n\n",
                lp$e$C[which.min(lp$flci$hl/lp$e$hl)]))
}

#' Replicate full analysis, using saved results if available
#' @param save.prefix Prefix for saving results of analysis as .RData file
#' @param fig.prefix Prefix for saving figures as .tex file
#' @param Cs vector of smoothness parameters
#' @param ylab y-label for y-axis in plots
#' @param C.ticks indices to C to display on x-axis of graphs
#' @param quiet print Empirical Analysis results?
#' @inheritParams RDHonest::RDHonest.fit
#' @export
FullAnalysis <- function(d, Cs, save.prefix, fig.prefix=NULL, se.initial="IKEHW",
                         ylab="", C.ticks=NULL, kern="triangular", se.method, quiet=TRUE) {
    ## don't use scientific notation
    scipenInitial <- getOption("scipen")
    options(scipen=7)
    on.exit(options(scipen=scipenInitial), add=TRUE)

    ptm <- proc.time()

    ## 0. Preliminary variance estimate
    d <- RDHonest::RDPrelimVar(d, se.initial)

    ## 1. Calculation
    savefile <- paste0(save.prefix, "-", min(Cs), "-", max(Cs), "-", length(Cs),
                       "-", se.initial, "-", se.method, ".RData")

    if (file.exists(savefile)) {
        load(savefile)
    } else {
        dopt <- tbl_RDTOpt(d, Cs, se.method)
        d1 <- tbl_RDLP(d, Cs, kern, "T", FALSE, 1, se.method)
        dh1 <- tbl_RDLP(d, Cs, kern, "H", FALSE, 1, se.method)
        d2 <- tbl_RDLP(d, Cs, kern, "T", FALSE, 2, se.method)
        dh2 <- tbl_RDLP(d, Cs, kern, "H", FALSE, 2, se.method)
        ## Single bw
        de1 <- tbl_RDLP(d, Cs, kern, "T", TRUE, 1, se.method)
        deh1 <- tbl_RDLP(d, Cs, kern, "H", TRUE, 1, se.method)
        save(file=savefile,
             dopt, d1, dh1, d2, dh2, de1, deh1, Cs, C.ticks,
             se.method, se.initial)
    }

    cat("Elapsed time: \n")
    print(proc.time()-ptm)

    ## 2. Display results
    if (!quiet) {
        RDEmpiricalAnalysis(d, d1, dopt)
        RDEmpiricalAnalysis(d, d2, dopt[1:3], "LQ")
        RDEmpiricalAnalysis(d, de1, dopt[1:3], "LL (equal bw)")
        RDEmpiricalAnalysis(d, dh1, lp.type="LL (Holder)")
        RDEmpiricalAnalysis(d, dh2, dh1, lp.type="LQ (Holder)",
                            opt.type="LL (Holder)")
        RDEmpiricalAnalysis(d, d1, dh1, lp.type="LL (Taylor)",
                            opt.type="LL (Holder)")
        RDEmpiricalAnalysis(d, deh1, dh1, lp.type="LL (equal bw, Holder)",
                            opt.type="LL (Holder)")
    }
    p2 <- plot_estimates(dopt, ylab=ylab, xticks=C.ticks, labels=FALSE)
    p3 <- plot_estimates(list(oci=dopt$e, e=dopt$e, flci=dopt$e),
                               ylab="", xticks=C.ticks)
    p4 <- plot_estimates(d1, ylab=ylab, xticks=C.ticks, labels=FALSE)
    p5 <- plot_estimates(list(oci=d1$e, e=d1$e, flci=d1$e),
                               ylab="", xticks=C.ticks)
    ## p6 <- plot_estimates(dh1, ylab=ylab, xticks=C.ticks)
    ## p7 <- plot_estimates(list(oci=dh1$e, e=dh1$e, flci=dh1$e),
    ##                            ylab=ylab, xticks=C.ticks)

    if (!requireNamespace("ggplot2", quietly = TRUE)) {
        stop("Please install 'tikzDevice' package;
              it's needed printing figures to .tex",
             call. = FALSE)
    }

    printFigures <- function(figfile, p) {
        ## 1511.06028v2. 5.2,4.2. With fullpage, we have letterpaper 8.5inches
        ## wide, with geometry page margins are 1in, so 3.2 width should fit two
        ## figures.
        tikzDevice::tikz(paste0(figfile, ".tex"), width=5.5, height=3.5,
                         verbose=FALSE)
        print(p)
        grDevices::dev.off()
        tikzDevice::tikz(paste0(figfile, "-slides.tex"), width=4.4, height=3.1,
                         verbose=FALSE)
        print(p)
        grDevices::dev.off()
    }

    if (!is.null(fig.prefix)) {
        figfile <- paste0(fig.prefix, "-", min(Cs), "-", max(Cs), "-",
                          length(Cs), "-", se.initial, "-", se.method)

        printFigures(paste0(figfile, "-opt"), p2)
        printFigures(paste0(figfile, "-opt-estimation"), p3)
        printFigures(paste0(figfile, "-ll"), p4)
        printFigures(paste0(figfile, "-ll-estimation"), p5)
        ## printFigures(paste0(figfile, "-llh"), p6)
        ## printFigures(paste0(figfile, "-llh-estimation"), p7)
    }

    if (se.method != "supplied.var") {
        bw <- RDHonest::IKBW.fit(d)
        cat("\nResults for Appendix C.1 of \"Optimal Inference\" paper\n")
        cat("\nIK bw:", bw, "\n")
        RDAltCI(d, kern, 1, bw, sclass="T", se.method=se.method)
        RDAltCI(d, kern, 2, bw, sclass="T", se.method=se.method)
    }

    ## RDAltCI(d, kern, 1, bw, sclass="H", se.method=se.method)
    ## RDAltCI(d, kern, 2, bw, sclass="H", se.method=se.method)

    ## Let's not report results at CCT bw
    ## cct <- rdrobust::rdbwselect(y=c(d$Ym, d$Yp),
    ##                             x=c(d$Xm, d$Xp))$bws[1, "h (left)"]
    ## cat("\nCCT bw:", cct, "\n")
    ## RDAltCI(d, kern, 1, cct, sclass="T", se.method=se.method)
    ## RDAltCI(d, kern, 2, cct, sclass="T", se.method=se.method)
    ## RDAltCI(d, kern, 1, cct, sclass="H", se.method=se.method)
    ## RDAltCI(d, kern, 2, cct, sclass="H", se.method=se.method)

    list(fig=list(opt=p2, opt.e=p3, ll=p4, ll.e=p5),
         dopt=dopt, d1=d1, d2=d2, dh1=dh1, dh2=dh2, Cs=Cs,
         C.ticks=C.ticks, data=d)
}

#' Replicate empirical analysis in Honest paper
#'
#' @param Ms vector of smoothness parameters M, such that the second derivative
#'     is bounded by M (i.e. \eqn{M=2C} in finite-sample paper)
#' @param bws bandwidths
#' @inheritParams RDHonest::RDHonest.fit
#' @export
HolderAnalysis <- function(d, kern, Ms=NULL, bws=NULL, se.method="nn",
                           J=3, alpha=0.05) {
    fs <- data.frame()
    d <- RDHonest::RDPrelimVar(d, se.initial="IKEHW")

    ## Finite-sample optimal bw given M
    rdh <- function(M, ...)
        RDHonest::RDHonest.fit(d, M, kern, sclass="H",
                               J=J, alpha=alpha, ...)

    if (!is.null(bws)) {
        for (b in bws) {
            ## Finite-sample method. For uniform kernel, objective is piecewise
            ## constant
            Est <- function(M) rdh(M, opt.criterion="MSE", bw.equal=TRUE,
                                   se.method="supplied.var")$hp-b
            M2 <- FindZero(Est, negative=FALSE)
            fs <- rbind(fs, as.data.frame(c(rdh(M2, hp=b,
                                                se.method=se.method)[-2],
                                            list(M=M2))))
        }
    } else if (!is.null(Ms)) {
        for (M in Ms) {
            Est <- rdh(M, opt.criterion="MSE", bw.equal=TRUE,
                       se.method=se.method)[-2]
            fs <- rbind(fs, as.data.frame(c(Est, list(M=M))))
        }
    }
    fs$cvb <- RDHonest::CVb(fs$maxbias/fs$sd, alpha)$cv

    fs
}
