#' Coverage of alternative confidence intervals
#'
#' Calculate implicit smoothness constant \code{C} such that naive local linear
#' or local quadratic CIs undercover by at most 5%, and compare these intervals
#' to optimal CIs
#' @inheritParams RDHonest::RDHonest.fit
#' @param hp,hm bandwidth for treated (units with positive running variable),
#'     and control (units with negative running variable) units.
#' @export
RDAltCI <- function(d, kern="triangular", order=1, hp, hm=hp, alpha=0.05,
                    sclass="T", se.method, se.initial) {

    uc <- 0.05                          # undercoverage tolerance
    rdh <- function(C, alpha, order, ...)
        RDHonest::RDHonest.fit(d=d, M=2*C, kern=kern, order=order, alpha=alpha,
                               se.method="supplied.var", sclass=sclass,
                               se.initial=se.initial, ...)
    undercover <- function(C)
        rdh(C, alpha, order, hp=hp, hm=hm)$naive-(1-alpha-uc)
    C1 <- FindZero(undercover, negative=FALSE)
    r1 <- rdh(C1, alpha, order, hp=hp, hm=hm)
    z <- qnorm(1-alpha/2)
    r2 <- rdh(C1, alpha, 1, opt.criterion="FLCI")
    r3 <- rdh(C1, alpha+uc, 1, opt.criterion="FLCI")
    CItype <- if (order==1) "Conventional" else "CCT"
    cat(CItype, " CI under ", sclass,
        sprintf("class:\nC=%f, bw=%f, CI: %f+-%f, coverage=%f",
                C1, hp, r1$estimate, z*r1$sd, (1-alpha-uc)),
        sprintf("\nHonest CI at this bw and C: %f+-%f, coverage=%f",
                r1$estimate, r1$hl, (1-alpha)),
        sprintf("\nOptimal LL CI at this C: %f+-%f, coverage=%f",
                r2$estimate, r2$hl, (1-alpha)),
        sprintf("\nOptimal LL CI at this C: %f+-%f, coverage=%f",
                r3$estimate, r3$hl, (1-alpha-uc)), sep="")
    if (sclass=="T") {
        r4 <- RDHonest::RDTOpt.fit(d, 2*C1, opt.criterion="FLCI",
                                   alpha=alpha, se.method=se.method,
                                   se.initial=se.initial)
        r5 <- RDHonest::RDTOpt.fit(d, 2*C1, opt.criterion="FLCI",
                                   alpha=alpha+uc, se.method=se.method,
                                   se.initial=se.initial)
        cat(sprintf("\nOptimal FLCI at this C: %f+-%f, coverage=%f",
                r4$estimate, r4$hl, (1-alpha)),
            sprintf("\nOptimal FLCI at this C: %f+-%f, coverage=%f",
                    r5$estimate, r5$hl, (1-alpha-uc)), sep="")
    }
    cat("\n\n")

    C1
}

#' Asymptotitcally optimal bw under Holder_{2}(M=2*C) class.
#'
#' Plug-in method for estimating asymtotically RMSE-optimal bandwidth for sharp
#' RD based on local linear regression.
#' @param C Smoothness constant. Second derivative is bounded by \code{2*C}
#' @param estimate.sigma Estimate sigma^2_+(0) and sigma^2_-(0), or use
#' \code{mean(d$sigma2p)} and \code{mean(d$sigma2m)} ?
#' @inheritParams RDHonest::IKBW.fit
#' @export
RMSEBandwidth <- function(d, kern="triangular", C, verbose=FALSE,
                          estimate.sigma=TRUE) {
    X <- c(d$Xm, d$Xp)
    Nm <- length(d$Xm)
    Np <- length(d$Xp)
    N <- Nm+Np

    ## STEP 0: Kernel constant, same as in RDHonest::IKBW.fit
    if (is.character(kern)) {
        s <- RDHonest::kernC[with(RDHonest::kernC,
                                order==1 & boundary==TRUE & kernel==kern, ), ]
    } else if (is.function(kern)) {
        ke <- RDHonest::EqKern(kern, boundary=TRUE, order=1)
        s <- list(mu2=RDHonest::KernMoment(ke, moment=2, boundary=TRUE, "raw"),
                  nu0=RDHonest::KernMoment(ke, moment=0, boundary=TRUE, "raw2"))
    }

    ## STEP 1: Estimate f(0), sigma^2_(0) and sigma^2_+(0)
    ## Silverman pilot bandwidth for uniform kernel
    h1 <- 1.84*stats::sd(X)/N^(1/5)
    f0 <- sum(X >= -h1 & X<= h1) / (2*N*h1)

    if (estimate.sigma==TRUE) {
        r1 <- RDHonest::RDLPreg(d=d, hp=h1, kern=kern, order=1,
                               se.method="nn")
        varm <- mean(r1$sigma2m)
        varp <- mean(r1$sigma2p)
    } else {
        varm <- mean(d$sigma2m)
        varp <- mean(d$sigma2p)
    }

    if(verbose)
        cat("\n h1: ", h1, "\n N_{-}, N_{+}: ", Nm, Np, "\n f(0): ", f0,
            "\n sigma^2_{+}(0): ", sqrt(varp),
            "^2\n sigma^2_{+}(0):", sqrt(varm), "^2\n\n")

    ## Final bandwidth: Equation (17)
    list(bw=unname(s$nu0/s$mu2^2 * (varp+varm) / (f0*N * 4*(2*C)^2))^(1/5),
         sigma2p=varp, sigma2m=varm, f0=f0)
}
