#' Simulation design
#' @param design one of 2, 3, 4, 5, 6, 7
#' @param sigma \eqn{sd(Y|X)}
#' @param n number of observations
#' @param M bound on second derivative of f(X)
#' @param b1,b2 knot for design 4
#' @param unif Uniform or beta distribution for X?
#' @importFrom stats runif rbeta rnorm
#' @keywords internal
RD_DGP <- function(design, sigma, n, M, b1, b2, unif) {

    X <- 2 * sort(if (unif) runif(n) else rbeta(n=n, shape1=2, shape2=4)) -1
    f <- function(X) {
        if (design==2 | design <= -5) {
            X*0
        } else if (design==3) {
            M/2 * (X^2*(X>=0) - X^2*(X<0))
        } else if (design==4 | design >=5 ) {
            M/2 * ((X^2-2*pmax(X-b1, 0)^2+2*pmax(X-b2, 0)^2)*(X>=0)-
                   (X^2-2*pmax(abs(X)-b1, 0)^2+2*pmax(abs(X)-b2, 0)^2)*(X<0))
        }
    }
    eps <- if (abs(design)<5) {
               sigma*rnorm(n)
           } else if (abs(design)==5) {
               sigma*(1+abs(X))*rnorm(n)
           } else if (abs(design)==6) {
               sigma*(1-abs(X))*rnorm(n)
           } else if (abs(design)==7) {
               sigma*(exp(rnorm(n))-exp(1/2))/sqrt(exp(1)*(exp(1)-1))
           }

    list(d=RDHonest::RDData(data.frame(y=f(X) + eps,
                                       x=X), cutoff=0), f=f)
}

#' Compute confidence intervals and estimates using a variety of methods
#' @param d RDData
#' @param C vector of values of C for Holder-honest CIs
#' @param what One of \code{"MSE"}, \code{"FLCI"}, or \code{"OCI"} (for
#' finite-sample optimal bws).
#' @keywords internal
getStats <- function(d, C, what="MSE") {
    bw <- RDHonest::IKBW.fit(d)
    ## Naive and CCT at IK bw
    ll <- RDHonest::RDLPreg(d, bw, order=1, se.method=c("EHW", "nn"), J=3)
    lq <- RDHonest::RDLPreg(d, bw, order=2, se.method=c("EHW", "nn"), J=3)
    Y <- c(d$Ym, d$Yp)
    X <- c(d$Xm, d$Xp)
    ## Naive and CCT at CCT defaults
    if (!requireNamespace("rdrobust", quietly = TRUE)) {
        stop("Please install rdrobust package;
              it's needed performing Monte Carlo.",
             call. = FALSE)
    }

    cct.ew <- rdrobust::rdrobust(y=Y, x=X, all=TRUE, vce="hc0")
    cct.nn <- rdrobust::rdrobust(y=Y, x=X, all=TRUE)

    r <- data.frame(method=c(rep(c("Naive", "RBC"), 2)),
                    bw=c(rep("IK", 2), rep("CCT", 2)),
                    bias=c(ll$estimate, lq$estimate,
                           unname(drop(cct.ew$coef)[c(1, 3)])),
                    se.ew=unname(c(ll$se["EHW"], lq$se["EHW"],
                                   drop(cct.ew$se)[c(1, 3)])),
                    se.nn=unname(c(ll$se["nn"], lq$se["nn"],
                                   drop(cct.nn$se)[c(1, 3)])),
                    h=unname(c(rep(bw, 2), rep(cct.ew$bws[1, 1], 2))),
                    rho=c(NA, 1, NA, cct.ew$bws[1, 1]/cct.ew$bws[2, 1]))
    za <- qnorm(0.975)
    r$hl.ew <- r$se.ew[1:4]*za
    r$hl.nn <- r$se.nn[1:4]*za
    r$r.ew <- abs(c((r$bias/r$se.ew)[1:2], cct.ew$z[c(1, 3)]))>za
    r$r.nn <- abs(c((r$bias/r$se.nn)[1:2], cct.nn$z[c(1, 3)]))>za

    ## Honest inference: first estimate sigma
    d <- RDHonest::RDPrelimVar(d, se.initial="IKEHW")
    ## sg <- RMSEBandwidth(d, C=1, verbose=FALSE, estimate.sigma=TRUE)
    ## d$sigma2p <- rep(mean(sg$sigma2p), length(d$Xp))
    ## d$sigma2m <- rep(mean(sg$sigma2m), length(d$Xm))

    for (Cj in C) {
        h.ew <- RDHonest::RDHonest.fit(d, M=2*Cj, sclass="H",
                                       opt.criterion=what,
                                       bw.equal=TRUE, se.method="EHW", J=3)
        h.nn <- RDHonest::RDHonest.fit(d, M=2*Cj, sclass="H", hp=h.ew$hp,
                             bw.equal=TRUE, se.method="nn", J=3)
        r <- rbind(r, data.frame(method=paste0("Hon, C=", Cj),
                                 bw=what, bias=h.ew$estimate,
                                 se.ew=h.ew$sd, se.nn=h.nn$sd, h=h.ew$hp,
                                 rho=NA, hl.ew=h.ew$hl, hl.nn=h.nn$hl,
                                 r.ew=abs(h.ew$estimate)>h.ew$hl,
                                 r.nn=abs(h.ew$estimate)>h.nn$hl))
    }
    rownames(r) <- NULL
    r
}

#' Monte Carlo simulation for coverage and length of CIs.
#'
#' Simulate CI coverage and length, comparing pointwise and uniform methods.
#' @param what One of \code{"Estimation"}, \code{"FLCI"}, or \code{"OCI"} (for
#'    selecting bw of Honest CIs).
#' @param Mh vector of values of M for Holder-honest CIs
#' @param nsim number of simulation draws
#' @inheritParams RD_DGP
RDSimulateSingle <- function(design, sigma, n, M, b1, b2, unif,
                             nsim, what="FLCI", Mh) {
    ## get dims of output
    se <- getStats(RD_DGP(design, sigma, n, M, b1, b2, unif)$d, Mh/2, what)
    r <- array(dim=c(nrow(se), ncol(se)-2, nsim))
    set.seed(42)
    for (j in 1:nsim)
        r[, , j] <- as.matrix(getStats(RD_DGP(design, sigma, n, M,
                                              b1, b2, unif)$d,
                                       Mh/2, what)[, -c(1, 2)])

    d <- cbind(se[, 1:2], rowMeans(r, dims=2))
    names(d) <- names(se)
    d
}


#' Monte Carlo simulation for coverage and length of CIs for a set of designs
#'
#' Simulate CI coverage and length, comparing pointwise and uniform methods
#' using a number of designs
#' @param p Data frame with design parameters.
#' @param what One of \code{"Estimation"}, \code{"FLCI"}, or \code{"OCI"} (for
#'    selecting bw of Honest CIs).
#' @param Mh vector of values of M for Holder-honest CIs
#' @param nsim number of simulation draws
#' @param eff.row Half-legnth comparisons should be wrt which row?
#' @inheritParams RD_DGP
#' @export
RDSimulate <- function(design=NULL, sigma=NULL, n=NULL, M=NULL,
                b1=NULL, b2=NULL, unif=NULL, p=NULL, nsim, what, Mh,
                eff.row=1) {
    if (is.null(p))
        p <- expand.grid(design=design, M=M, sigma=sigma, n=n, b1=b1,
                         b2=b2, unif=unif)
    r <- list()
    s <- Sys.time()
    for (j in seq_len(nrow(p))) {
        rj <- RDSimulateSingle(design=p$design[j], sigma=p$sigma[j],
                               n=p$n[j], M=p$M[j], b1=p$b1[j], b2=p$b2[j],
                               unif=p$unif[j], nsim=nsim, what=what, Mh=Mh)
        rj$eff.ew <- rj$hl.ew/rj$hl.ew[eff.row]
        rj$eff.nn <- rj$hl.nn/rj$hl.nn[eff.row]

        r <- rbind(r, cbind(as.data.frame(lapply(p[j, ], function(a)
            rep(a, nrow(rj)))), rj))

        print(p[j, ])
        cols <- c("method", "bw", "bias", "eff.ew", "eff.nn",
                  "h", "rho", "r.ew", "r.nn")
        pander::pandoc.table(rj[, cols], round=3, style="simple")
    }
    print(difftime(Sys.time(), s))

    r
}

#' Plot functions used in Monte Carlo design 4
#'
#' The parameters \code{b1}, \code{b2} should be same length, and \code{M}
#' should be scalar
#' @inheritParams RD_DGP
#' @export
plot_design <- function(M, b1, b2) {
    X <- seq(0, 1, by=0.005)
    Y <- sapply(1:length(b1), function(j)
        RD_DGP(design=4, sigma=1, n=100, M=M, b1[j], b2[j], TRUE)$f(X))
    w <- sapply(1:length(b1), function(j)
        paste0("$b_{1}=", b1[j], "$, $b_{2}=", b2[j], "$"))
    d <- data.frame(x=rep(X, length(b1)), y=as.vector(Y),
                    what=rep(w, each=length(X)))

    baseplot(d, ylab="$f_{+}(x)$", xlab="$x$",
             method=list(cex=0.9, "last.bumpup"))
}
