## Bounded normal mean model

#' Compute optimal affine fixed-length CI in bounded normal mean model
#'
#' Given an observation \eqn{X\sim N(\theta, 1)}, with
#' \eqn{\theta\in[-\tau,\tau]}, compute optimal affine fixed-length confidence
#' interval for \eqn{theta}, \eqn{cX\pm \chi}, with coverage \eqn{1-\alpha}
#' @param tau Bound on parameter space, non-negative number.
#' @param alpha Determines CI level, \code{1-alpha}.
#' @return list with two elements, \code{c} and \code{chi}. Optimal CI is given
#' by has lower and upper limits given by \code{c*X-chi} and \code{c*X+chi},
#' respectively
#' @examples
#' bnmflci(tau = 3, alpha = 0.1)
#' @export
bnmflci <- function(tau, alpha=0.05) {
    ## In this case we can't shrink
    if (tau==Inf) return(list(c=1, chi=qnorm(1-alpha/2)))

    ## First, given chi, find optimal shrinkage c_{0-1,chi}X that maximizes
    ## coverage of the interval cX\pm chi. Next function computes 1/c_{0-1,chi}.
    ## It assumes chi<tau, otherwise optimal c=0
    cinv <- function(chi)
        if (chi >= tau ) {
            Inf
        } else  {
            1/2+sqrt(1/4+log((tau+chi)/(tau-chi))/(2*tau*chi))
        }

    ## resulting coverage, see Appendix B in the paper
    cover <- function(chi)
        if (chi >= tau) {
            1
        } else {
            stats::pnorm(cinv(chi)*(tau+chi)-tau) -
                stats::pnorm(cinv(chi)*(tau-chi)-tau)
        }

    ## Optimal length: set coverage to 1-alpha
    chi <- if (tau < qnorm(1-alpha)) {
        tau
    } else {
        stats::uniroot(function(chi) cover(chi)-(1-alpha), c(1e-8, tau),
                       tol=tol)$root
    }

    ## shrinkage at optimal length
    list(c=1/cinv(chi), chi=chi)
}

#' Plot optimal shrinkage and optimal CI length in bounded normal mean model
#' @return two \code{ggplot2} plot objects.
#' @export
plot_bnm <- function() {
    ## optimal shrinkage for estimation vs CI
    taus <- seq(0.01, 5, by=0.01)
    lt <- length(taus)
    df <- t(cbind(sapply(taus, bnmflci, alpha=0.1),
                  sapply(taus, bnmflci, alpha=0.05),
                  sapply(taus, bnmflci, alpha=0.01),
                  rbind(taus^2/(1+taus^2), NA)))
    df <- data.frame(c=as.numeric(df[, "c"]), chi=as.numeric(df[, "chi"]))
    labels <- c(rep("$\\alpha=0.1$", lt), rep("$\\alpha=0.05$", lt),
               rep("$\\alpha=0.01$", lt), rep("MSE", lt))
    df <- cbind(x=taus, df, what=labels)

    br <- c(0, qnorm(0.9), qnorm(0.95), qnorm(0.99), 3: 5)
    br.lab <- c("0", "$z_{0.9}$", "$z_{0.95}$", "$z_{0.99}$", 3:5)

    p1 <- baseplot(data=cbind(df, data.frame(y=df$c)),
                   ylab="", xlab="$\\tau$", br=br, br.lab=br.lab,
                   xlim=c(-0.2, 5.5), method=list(cex=0.9, "lasso.labels")) +
        ggplot2::ylim(c(-0.1, 1))

    ## optimal CI length
    p2 <- baseplot(data=cbind(df, data.frame(y=df$chi)),
                   ylab="", xlab="$\\tau$", br=br, br.lab=br.lab,
                   xlim=c(-0.2, 5.5), method=list(cex=0.9, "last.qp"))
    list(p1=p1, p2=p2)
}
