## Compute relative efficiency of kernels under Taylor and Holder classes

#' Compute bias and variance constants for a given kernel and local polynomial
#' regression of order \code{order}
#' @inheritParams RDHonest::EqKern
#' @keywords internal
SYBiasVar <- function(kernel="uniform", order, boundary=TRUE,
                         equivalent=FALSE) {

    pis <- paste("pi", 1:3, sep="")
    if (!is.function(kernel)) {
        ## Get kernel roughtness R and absolute moments pi from table
        row <- RDHonest::kernC$kernel==kernel & RDHonest::kernC$order==order &
            RDHonest::kernC$boundary==boundary
        R <- RDHonest::kernC[row, "nu0"]
        pik <-as.numeric(RDHonest::kernC[row, pis])
    } else  {
        K <- if(equivalent) kernel else RDHonest::EqKern(kernel, boundary, order)
        mu0 <- RDHonest::KernMoment(K, moment = 0, boundary, "raw")
        R <- RDHonest::KernMoment(K, moment = 0, boundary, "raw2") / mu0
        pik <- sapply(1:3, function(j)
            RDHonest::KernMoment(K, j, boundary, "absolute") / mu0)
    }

    ## if order too low, const doesn't matter
    if (order<2)
        pik[(order+2):3] <- NA

    ## bw=optimal bandwidth constant
    list(R=R, B=pik, bw=R/pik[order+1]^2, b=rep(NA, 3))
}


#' Return function \eqn{s_{p,K}(b)=\int (|x|-b)_{+}^{p} k(x) dx}
#' @param K kernel function \eqn{k}
#' @param boundary If true, integrate over [0,1], otherwise integrate over [-1,1]
#' @keywords internal
s <- function(K, boundary=FALSE, p=1){
    c <- 1+ 1*(boundary==FALSE)
    function(b) c*stats::integrate(function(u) (u-b)^p * K(u),
                                   lower=b, upper=1, rel.tol=tol)$value
}

#' Compute bias and variance constants for a given kernel and local polynomial
#' regression of order \code{order}
#' @inheritParams RDHonest::EqKern
#' @keywords internal
HolBiasVar <- function(kernel="uniform", order, boundary=TRUE,
                          equivalent=FALSE) {
    p <- 1:3                            # order of Holder class
    K <- if(equivalent) kernel else RDHonest::EqKern(kernel, boundary, order)

    mu <- function(p) {
        lower <- if (boundary==TRUE) 0 else -1
        stats::integrate(function(u) abs(u)^p*K(u),
                         lower, 1, rel.tol=tol)$value
    }

    ## Optimize s(b)
    sOpt <- function(p, max)
        stats::optimize(s(K, boundary, p=p), c(0, 1), maximum=max, tol = tol)

    if (!is.function(kernel)) {
        ## Get kernel roughtness R and absolute moments pi from table
        row <- RDHonest::kernC$kernel==kernel & RDHonest::kernC$order==order &
            RDHonest::kernC$boundary==boundary
        R <- RDHonest::kernC[row, "nu0"]
    } else  {
        mu0 <- RDHonest::KernMoment(K, moment = 0, TRUE, "raw")
        R <- RDHonest::KernMoment(K, moment = 0, TRUE, "raw2") / mu0
    }

    ## Compute optimal b's, points at which the least favorable function changes
    ## direction, and maximum bias B
    B <- rep(NA, length(p))
    b <- rep(NA, 3)

    if (order==0) {
        B[1] <- mu(1)
    } else if (order==1 & boundary==FALSE) {
        B[1:2] <- c(mu(1), mu(2))
    } else if (order==1 & boundary==TRUE) {
        r <- sOpt(p=1, FALSE) # find minimum
        b[1] <- r$minimum
        B[1:2] <- c(-2*r$objective, -mu(2))
    } else if (boundary==TRUE & order==2) {
        r1 <- sOpt(p=1, TRUE)
        r2 <- sOpt(p=1, FALSE)
        r3 <- sOpt(p=2, TRUE)
        stopifnot(sOpt(p=3, TRUE)$maximum<1e-5)
        stopifnot(sOpt(p=3, FALSE)$minimum>1-1e-5)

        b <- c(r2$minimum, r1$maximum, r3$maximum)
        B <- c(2*(r1$objective-r2$objective), 2*r3$objective, mu(3))
    } else if (order==2 & boundary==FALSE) {
        r1 <- sOpt(p=1, FALSE)
        r2 <- sOpt(p=2, FALSE)
        stopifnot(sOpt(p=3, FALSE)$minimum<1e-5)

        b[c(1, 3)] <- c(r1$minimum, r2$minimum)
        B <- c(mu(1)-2*r1$objective, -2*r2$objective, -mu(3))
    }

    ## bw=optimal bandwidth constant
    list(R=R, B=B, b=b, bw=R/B[order+1]^2)
}

#' Constant \eqn{R(k)^{r}\mathcal{B}(k)^{1-r}} for optimal kernel
#'
#' @param sclass one of \code{"T"} or \code{"H"}
#' @inheritParams RDHonest::EqKern
#' @keywords internal
optConst <- function(p, boundary=TRUE, sclass="T") {
    if (sclass=="T") {
        B <- RDHonest::KernMoment(SYKernel(p, boundary=boundary)$k,
                                p, boundary, "absolute")
        R <- RDHonest::KernMoment(SYKernel(p, boundary=boundary)$k,
                                0, boundary, "raw2")
        return(R^(p/(2*p+1)) * B^(1/(2*p+1)))
    } else {
        HolConstant(p, boundary=boundary)
    }
}



#' Relative efficiency of kernels under Holder or Taylor smoothness class
#'
#' Compute relative efficiency of using a given kernel \code{kernel} and local
#' polynomial regression of order \code{order} relative to using the optimal
#' kernel under Holder or Taylor smoothness class of order \code{p}, with
#' \code{p} equal to 1, 2, or 3.
#' @param equivalent Is the kernel function supplied an equivalent kernel?
#' @param sclass Determines smoothness class, one of \code{"T"} (for Taylor) or
#' \code{"H"} (for Holder)
#' @inheritParams RDHonest::EqKern
#' @return list with elements
#' \describe{
#'   \item{R}{\eqn{\int k^2}}
#'   \item{B}{vector of bias constants \eqn{\mathcal{B}_p}, for each order
#'            \eqn{p}, with some possibly \code{NA} if order is too low }
#'   \item{Eff}{Kernel efficiency for each order \eqn{p} of the class}
#'   \item{bw}{bandwidth constant \code{R/B[order+1]^2}}
#' }
#' @export
KerEfficiency <- function(kernel="uniform", order, boundary=TRUE,
                          equivalent=FALSE, sclass="H") {

    p <- 1:3                            # order of smoothness class

    ## constants for kernel in question
    C <- if (sclass=="H")
             HolBiasVar(kernel, order, boundary, equivalent)
         else
             SYBiasVar(kernel, order, boundary, equivalent)

    ## Relative efficiency
    eff <- sapply(p, optConst, boundary=boundary, sclass=sclass) /
        (C$R^(p/(2*p+1)) * C$B^(1/(2*p+1)))

    addp <- function(string) paste0(string, p)

    r <- c(C$R, C$B, eff, C$bw, C$b)

    bs <- c("b11", "b12", "b2")
    names(r) <- c("R", addp("B"), addp("Eff"), "bw", bs)
    r
}

#' Efficiency gains from imposing Holder class over Taylor class.
#' @return Data frame with efficiency gains.
#' @export
tbl_HolderGains <- function() {
    kernels <- c("uniform", "triangular", "epanechnikov")

    eff <- function(kernel, boundary, p) {
        h <- KerEfficiency(kernel, p-1, boundary, sclass="H")
        t <- KerEfficiency(kernel, p-1, boundary, sclass="T")

        unname((h["R"]/t["R"])^(p/(2*p+1)) *
                          (h[paste0("B", p)]/t[paste0("B", p)])^(1/(2*p+1)))
    }
    opt <- function(boundary, p)
        optConst(p, boundary, sclass="H") / optConst(p, boundary, sclass="T")
    gains <- function(bd)
        sapply(1:3, function(p) c(sapply(kernels, eff, bd, p), opt=opt(bd, p)))
    r <- cbind(gains(TRUE), gains(FALSE))
    colnames(r) <- c("Bd1", "Bd2", "Bd3", "Int1", "Int2", "Int3")
    r
}


#' Table of Relative efficiency of kernels under Taylor or Holder class.
#'
#' @inheritParams KerEfficiency
#' @return two data frames, one for interior, one for boundary
#' @export
tbl_KerEfficiency <- function(sclass="H") {
    effk <- function(kern, boundary)
        data.frame(kernel=kern, order=0:2,
                   t(sapply(0:2, function(j)
                       KerEfficiency(kernel=kern,  j, boundary=boundary,
                                     sclass=sclass))))

    kernels <- c("uniform", "triangular", "epanechnikov")
    bound <- do.call(rbind, c(lapply(kernels, function(k) effk(k, TRUE))))
    inter <- do.call(rbind, c(lapply(kernels, function(k) effk(k, FALSE))))

    list(interior=inter, boundary=bound)
}

## Performance of alternative confidence intervals

#' Compute coverage and relative length of Alternative confidence intervals
#'
#' Compute coverage and length of CCT and naive CIs (relative to honest CIs
#' around MSE optimal estimator). For CCT, the calculation assumes that the
#' pilot bandwidth and main bandwidth are both equal. The calculation further
#' assumes that the bandwidth is MSE-optimal.
#' @param sclass one of \code{"T"} or \code{"H"}
#' @param alpha Determines CI level, \code{1-alpha}.
#' @param ptwise is the MSE-optimal bandwidth optimal under uniform asymtptotics
#'     or pointwise asymptotics?
#' @param mult Multiply the MSE-optimal bandwidth by \code{mult}
#' @param what either "CCT" or "naive"
#' @inheritParams RDHonest::EqKern
#' @export
AltCIEfficiency <- function(kernel="uniform", boundary=TRUE, sclass="H",
                          alpha=0.05, ptwise=FALSE, what="CCT", mult=1) {
    ll <- KerEfficiency(kernel, order=1, boundary=boundary, sclass=sclass)
    lq <- KerEfficiency(kernel, order=2, boundary=boundary, sclass=sclass)
    ## for pointwise optimal bandwidth
    llp <- RDHonest::kernC[RDHonest::kernC$kernel==kernel &
                           RDHonest::kernC$order==1 &
                           RDHonest::kernC$boundary==boundary, ]

    R.ratio <- if (what=="naive") 1 else sqrt(ll["R"]/lq["R"])
    B0 <- if (ptwise) abs(llp$mu2) else ll["B2"] # mathcal{B} used in bw
    B.ratio <- (if (what=="naive") ll["B2"] else lq["B2"]) / B0

    t <- 1/2 * R.ratio * B.ratio * (mult)^(5/2) # t_{q,p}
    z <- stats::qnorm(1-alpha/2)
    covr <- stats::pnorm(t+z)-stats::pnorm(t-z)
    rl <- (mult)^(-1/2)* z/RDHonest::CVb(1/2, alpha)$cv /
        R.ratio * (B0 / ll["B2"])^(1/5)
    list(relative.length=unname(rl), coverage=unname(covr), t=unname(t))
}

#' Table of coverate and relative length of alternative confidence intervals
#' @inheritParams AltCIEfficiency
#' @return Data frame of efficiencies
#' @export
tbl_AltCIEfficiency <- function(sclass="H", ptwise=FALSE, what="CCT",
                                mult=1) {
    kernels <- c("uniform", "triangular", "epanechnikov")
    rb <- as.numeric(t(sapply(kernels, AltCIEfficiency, sclass=sclass,
                              ptwise=ptwise, what=what, mult=mult)))
    ri <- as.numeric(t(sapply(kernels, AltCIEfficiency, boundary=FALSE,
                              sclass=sclass, ptwise=ptwise, what=what,
                              mult=mult)))
    ## Convert to standard data frame with columns numeric vectors

    data.frame(kernel=rep(kernels, 2),
               boundary=rep(c("Yes", "No"), each=3),
               relative.length=c(rb[1:3], ri[1:3]),
               coverage=c(rb[4:6], ri[4:6]),
               t=c(rb[7:9], ri[7:9]))
}
