## Optimal kernels under Holder class

#' Compute inverse modulus \eqn{\omega^{-1}(2,1)} for Holder(2) class
#'
#' @return List with two elements, inverse modulus \eqn{\omega^{-1}(2,1)}, and,
#' in the boundary case, \eqn{x_0} the point at which the derivative of the
#' least favorable function is zero.
#' @keywords internal
HolInverseM <- function(boundary=TRUE) {
    ## Analytic interior solution, first sign change at sqrt(1+q)
    q <- (3+sqrt(33)-sqrt(26+6*sqrt(33)))^2/16

    omegai <- sqrt(8 * sqrt(1+q) * (23*q^2-14*q+23) / (30*(1-q^(5/2))))
    x0 <- NA

    if (boundary) {
        obj <- function(x)
            x+2*x^5/15-2*x^3/3+abs(1-x^2/2)^(5/2)*omegai^2/8
        r <- stats::optimize(obj, c(1, 2), tol=1e-14)
        omegai <- 2*sqrt(r$objective)
        x0 <- r$minimum
    }

    return(list(omegai=omegai, x0=x0))
}

#' Compute the optimal kernel constant \eqn{\|k\|^{r}\mathcal{B}(k)^{1-r}}
#' @keywords internal
HolConstant <- function(p=1, boundary=TRUE) {
    if (p==1 | p==3) {
        ## Triangular kernel optimal for p=1, best we come up with for p=3
        opt <- HolBiasVar("triangular", p-1, boundary)
        const <- (opt$B[p])^(1/(2*p+1)) * (opt$R)^(p/(2*p+1))
    } else if (p==2) {
        omegai <- HolInverseM(boundary)$omegai
        const <- (2)^(1/5)* 4^(4/5)/5 *(2/omegai)^(4/5)
    }
    const
}

#' Optimal equivalent kernel in Holder(2) smoothness class.
#'
#' Compute the optimal equivalent kernel for Holder(2) smoothness class.
#' Parameter of interest is value of \eqn{f} at a point.
#'
#' The kernel is normalized to be supported on [-1, 1] (for interior
#' regression), or [0, 1] (for boundary regression)
#' @param boundary logical scalar, specifying whether point of interest is at
#' boundary of parameter space.
#' @return a list with names \code{c("k", "supp")}, the first element is a
#' function that corresponds to the optimal kernel, the second element is the
#' support endpoint of the least favorable function \eqn{f_{2,1}}. The integral
#' of \eqn{f_{2,1}} is given by \code{supp/k(0)}
#' @examples
#' HolKernel(TRUE)$k(0.5)
#' @export
HolKernel <- function(boundary=TRUE) {

    ## Analytic interior solution, first sign change at sqrt(1+q)
    q <- (3+sqrt(33)-sqrt(26+6*sqrt(33)))^2/16
    kmax <- 2*sqrt(1+q)/(1-sqrt(q))     # support

    f <- function(x) {
        ## Sign changes
        k <- sqrt(1+q)/(1-sqrt(q))*(2-q^((0:10)/2)-q^((1:11)/2))
        sc <- function(j) (-1)^(j+1)*pmax(abs(x)-k[j], 0)^2
        (1 - x^2/2 + Reduce("+", lapply(1:10, sc)))*(abs(x)<kmax)
    }

    if (!boundary) {
        ## rescale so it integrates to one and has support on [-1,1]
        r <- stats::integrate(function(x) f(x*kmax), 0, 1, rel.tol=tol)$value
        return(list(k=function(x) f(x*kmax)/(2*r), supp=kmax))
    } else {
        x0 <- HolInverseM(TRUE)$x0
        b <- abs(1-x0^2/2)
        fb <- function(x) (x>=0)*(x<=x0)*(1 + x^2/2 - x*x0)-
            (x>x0)*b*f((x-x0)/sqrt(b))
        kbmax <- x0+sqrt(b)*kmax       # support

        ## rescale so it integrates to one and has support on [0,1]
        r <- stats::integrate(function(x) fb(x*kbmax), 0, 1, rel.tol=tol)$value
        return(list(k=function(x) fb(x*kbmax) / r, supp=kbmax))
    }
}

#' Plot optimal kernels under Holder(2) smoothness class
#' @inheritParams HolKernel
#' @return ggplot2 object
#' @export
plot_HolKernel <- function(boundary=FALSE) {
    us <- seq(if (boundary) 0 else -1, 1, length.out=1000)
    ks <- HolKernel(boundary=boundary)$k(us)
    df <- data.frame(x=rep(us, 1), y=as.vector(ks),
                     what=rep("$p=2$", each=length(us)))

    baseplot(df, "$k^*(u)$", "$u$")
}
