## Optimal kernels under Taylor smoothness classes


#' basic form of optimal SY kernel
#'
#' \eqn{k(x)=(b+d'r(x)-Cx^p)_{+} - (b+d'r(x)+Cx^p)_{-}}, where
#' \eqn{r(x)=(x,\dotsc,x^{p-1})}
#' @keywords internal
SYBasic <- function(x, b=1, d=NULL, C=1) {
    p <- length(d) + 1
    dx <- if (p > 1) drop(sapply(seq_len(p-1), function(j) x^j) %*% d) else 0
    pmax(b+dx-C*abs(x)^p, 0) + pmin(b+dx+C*abs(x)^p, 0)
}


#' support of optimal SY kernel
#' @keywords internal
SYsupp <- function(d, b=1, C=1, boundary=FALSE) {
    roots <- c(polyroot(c(b, d, C)), polyroot(c(b, d, -C)))
    ## biggest real root
    if (boundary==FALSE) {
        max(abs(Re(roots[Im(roots)<tol])))
    } else {
        max(Re(roots[Im(roots)<tol]))
    }
}


#' Solve for for \eqn{d} given \eqn{b=1}, and \eqn{C=1}.
#' @param p Order of Sacks-Ylvisacker class, integer greater than one.
#' @param start Starting value for \code{d}.
#' @param boundary Logical scalar, specifying whether we are at a boundary.
#' @param verbose Logical scalar, should rootSove be verbose?
#' @keywords internal
SYfindd <- function(p=2, boundary=FALSE, start=NULL, verbose=FALSE) {
    if (p==1) {
        return(NULL)
    } else if (p==2 && boundary==TRUE) {
        ## Closed-form expression for \int u*k(u) available
        int2 <- function(c)
            (2*c*(c^2-4)^(3/2) +6 +6*c^2 + c^4 + c*(c^2+4)^(3/2)) / 24
        return(stats::uniroot(int2, interval=c(-2.1, -2), tol=tol)$root)
    } else if (p==2 && boundary==FALSE) {
        return(0)                       # Epanechnikov kernel
    } else {
        intp <- function(d) {
            s <- SYsupp(d, boundary=boundary)
            sapply(1:(p-1), function(j)
                stats::integrate(f=function(x) (x^j)*SYBasic(x, d=d),
                          lower= -s*(1-boundary), upper=s, subdivisions=1000,
                          rel.tol = tol)$value)
        }

        ## starting value
        st <- if (!is.null(start)) {
                  start
              } else if (p==3 && boundary==TRUE) {
                  c(-4.54028, 4.01576)
              } else if (p==3 && boundary==FALSE) {
                  c(0, -1.9)
              } else if (p==4 && boundary==FALSE) {
                  c(0, -2.1339, 0)
              } else if (p==4 && boundary==TRUE){
                  c(-7.2, 13.6, -6.8)
              }

        return(rootSolve::multiroot(intp, start = st, verbose = verbose,
                                    rtol = tol)$root)
    }
}


#' Optimal equivalent kernel in Sacks-Ylvisacker smoothness class.
#'
#' Compute the optimal equivalent kernel for Sacks-Ylvisacker smoothness class,
#' \deqn{\{|f(x)-\sum_{j=0}^{p-1}f^(j)(0)x^j| <= C|x|^p/p!\}.} Parameter of
#' interest is value of \eqn{f} at a point.
#'
#' The kernel is normalized to be supported on [-1, 1] (for interior
#' regression), or [0, 1] (for boundary regression).
#' @param p Order of Sacks-Ylvisacker class, positive integer.
#' @param boundary Logical scalar, specifying whether we are at a boundary.
#' @param integral Positive number that the integral \eqn{\int_0^1 k} should
#'     integrate to. For boundary regression, the default is \code{1}, for
#'     interior regression the default is \code{1/2}.
#' @param recompute Should the kernel parameters \eqn{d}, \eqn{C}, and \eqn{b}
#'     be recomputed? Default is \code{FALSE}, in which case pre-computed values
#'     are used if \code{p} is smaller than 5.
#' @param start Starding value for computation of \code{d}, only needed if
#'     \code{recompute} is \code{TRUE}.
#' @return a list with names \code{c("k", "b", "d", "C")}, the first element is
#'     a function that corresponds to the optimal kernel, the remaining three
#'     elements are the kernel parameters.
#' @examples
#' SYKernel(2)
#' @export
SYKernel <- function(p, boundary=FALSE, integral=1/2+1/2*boundary,
                     recompute=FALSE, start=NULL) {

    ## The loaded values of b, d, C are such that the kernel integrates over [0,
    ## 1] to its default values.
    if (recompute==FALSE && p<=4){
        pr <- if(boundary==FALSE) SYKernelI[[p]] else SYKernelB[[p]]
        sc <- 1/2+1/2*boundary
    } else {
        d0 <- SYfindd(p=p, boundary=boundary, start=start)
        ## rescale so support is on [-1, 1]
        s0 <- SYsupp(d0, boundary=boundary)
        pr <- list(b = 1, d = d0*s0^seq_len(length(d0)),
                       C = s0^(length(d0)+1))

        ## compute integral
        sc <- stats::integrate(function (x) SYBasic(x, b=pr$b, d=pr$d, C=pr$C),
                               lower=0, upper=1, rel.tol = tol)$value
    }
    pr <- lapply(pr, function(j) integral/sc * j)

    c(list(k=function(x) SYBasic(x, b=pr$b, d=pr$d, C=pr$C)), pr)
}


#' Plot optimal kernels under Taylor smoothness class.
#' @param p Order of Sacks-Ylvisacker class, positive integer.
#' @param boundary Boundary kernel?
#' @param facet Plot as a facet (grid)?
#' @return ggplot2 object with legend suppressed (so that it can be added
#'     directly onto the plot with the \code{directlabels} library)
#' @export
plot_SYKernel <- function(p=2:4, boundary=FALSE, facet=TRUE) {
    us <- seq(if (boundary) 0 else -1, 1, length.out=1000)
    ks <- sapply(p, function(or) SYKernel(p=or, boundary=boundary)$k(us))
    df <- data.frame(x=rep(us, length(p)), y=as.vector(ks),
                     what=rep(paste0("$p=", p, "$"), each=length(us)))

    baseplot(df, "$k^*(u)$", "$u$", facet=facet)
}
