## General results under renormalization

## 1. Efficiency bounds

#' Asymptotic efficiency bounds for minimax CIs
#'
#' Compute efficiency of minimax onesided-CIs relative to power-directed CIs, or
#' fixed-length CIs relative to Pratt CIs that direct all power at a constant
#' function. The efficiency bound assumes that the modulus renormalizes, so that
#' it depends only on the rate exponent \eqn{r}
#' @param r vector of rate exponents
#' @param what one of "FLCI", "OCI", or "CLFLCI" (Cai-Low bound for additional
#'     materials for referees)
#' @param alpha Determines CI level, \code{1-alpha}.
#' @return Scalar between 0 and 1
#' @importFrom stats qnorm
#' @export
AEffBoundCI <- function(r, alpha=0.05, what="FLCI") {
    if (what=="OCI") {
        return(2^(r)/(1+r))
    } else {
        ## c^{-1}_{\chi}(r)
        cinv <- function(r)
            FindZero(function(tau) bnmflci(tau, alpha=alpha)$c-r,
                     negative=FALSE)

        ## See Equation (22) in the paper, section 3.4
        eff <- function(r) {
            fk <- function(z) (qnorm(1-alpha) - z)^r * stats::dnorm(z)
            num <- if (what=="CLFLCI") {
                       (1/2-alpha) * qnorm(1-alpha)^r / 2
                   } else {
                       stats::integrate(fk, -10, qnorm(1-alpha),
                                        rel.tol=tol)$value
                   }
            den <- bnmflci(cinv(r), alpha=alpha)$chi * cinv(r)^(r-1)
            num / den
        }
        return(sapply(r, eff))
    }
}


#' Plot asymptotic efficiency bounds for minimax and fixed-length CIs
#' @return ggplot2 plot object.
#' @export
plot_AEffBoundCI <- function() {
    ## optimal shrinkage for estimation vs CI
    r <- seq(1/2, 0.95, by=0.01)

    df <- data.frame(x=rep(r, 2),
                     y=c(AEffBoundCI(r, what="OCI"),
                         AEffBoundCI(r, what="FLCI")),
                     what=rep(c("Minimax onesided", "Fixed-length"),
                              each=length(r)))

    baseplot(df, ylab="Relative efficiency", xlab="$r$",
             xlim=c(0.5, 1.05), facet=FALSE, method=list(cex=0.9, "lines2"))
}

## 2. Optimal inference based on kernels


#' Part of asymptotic risk that depends on bandwidth
#' @keywords internal
BWRisk <- function(r, t, what="FLCI", alpha=0.05, beta=0.5) {
    if (what=="FLCI") {
        t^(r-1) * RDHonest::CVb(t, alpha=alpha)$cv
    } else if (what=="OCI") {
        t^(r-1) * (2*t+qnorm(beta)+qnorm(1-alpha))
    } else if (what=="MSE") {
        t^(r-1) * sqrt(t^2+1)
    }
}

#' Compute \eqn{t^*}, optimal bias-variance trade-off.
#'
#' @param r Rate exponent
#' @param what what performance criterion to use. One of "FLCI" (fixed-length
#' CI), "OCI" (one-sided CI), "MSE" (mean-squared error)
#' @param alpha for "FLCI" and "OCI", determines confidence level,
#' \code{1-alpha}.
#' @param beta for "OCI", quantile of excess length to optimize
#' @keywords internal
tstar <- function(r, what="FLCI", alpha=0.05, beta=0.5)
    if (what=="FLCI") {
        stats::optimize(function(t) BWRisk(r, t, what="FLCI", alpha=alpha),
                        interval=c(0, 3))$minimum
    } else if (what=="OCI") {
        (1/r-1)*(qnorm(beta)+qnorm(1-alpha))/2
    } else if (what=="MSE") {
        sqrt(1/r-1)
    }


#' Compute optimal bias-sd ratio \eqn{t^*} and MSE-bandwidth efficiency
#'
#' Computes optimal \eqn{t^*} for MSE, OCI, and FLCI, as well as relative
#' efficiency of confidence intervals centered around MSE-optimal bandwidth.
#'
#' @param r rate exponent (optimal rate of convergence). Can be a vector
#' @param beta For "OCI", quantiles of excess length to optimize. Can be a
#'     vector.
#' @param alpha Determines CI level for "FLCI" and "OCI", \code{1-alpha}. Can be
#'     a vector.
#' @return Returns data frame with following columns
#' \describe{
#'     \item{r}{Rate exponent as specified by \code{r}}
#'     \item{alpha}{Value of \eqn{\alpha} as specified by \code{alpha}}
#'     \item{beta}{Value of \eqn{\beta} as specified by \code{beta}}
#'     \item{criterion}{Efficiency criterion, MSE, OCI or FLCI}
#'     \item{t}{Optimal maximum bias-sd ratio, \eqn{t^*}}
#'     \item{MSEEfficiency}{Efficiency of minimax MSE-optimal bandwidth at this criterion}
#'     \item{TeXDescription}{LaTeX-friendly description of current row}
#' }
#' @examples
#' BWEfficiency(r=c(1/2, 2/3, 4/5), beta=0.5, alpha=c(0.1, 0.05))
#' @export
BWEfficiency <- function(r, beta, alpha) {

    ## Optimal t
    mse <- data.frame(r=r, alpha=NA, beta=NA, criterion="MSE",
                      t=sapply(r, function(r) tstar(r, what="MSE")),
                      MSEEfficiency=1)
    fl <- expand.grid(r=r, alpha=alpha, beta=NA, criterion="FLCI")
    fl$t <- apply(fl[, -4], 1, function(u)
        tstar(r=u["r"], what="FLCI", alpha=u["alpha"]))
    os <- expand.grid(r=r, alpha=alpha, beta=beta, criterion="OCI")
    os$t <- apply(os[, -4], 1, function(u)
        tstar(r=u["r"], what="OCI", alpha=u["alpha"], beta=u["beta"]))

    ## Efficiency of MSE
    eff <- function(u, what) {
        BWRisk(u["r"], u["t"], what=what, alpha=u["alpha"], beta=u["beta"]) /
            BWRisk(u["r"], mse[mse$r==u["r"], "t"],
                   what=what, alpha=u["alpha"], beta=u["beta"])
    }

    fl$MSEEfficiency <- apply(fl[, -4], 1, function(u) eff(u, what="FLCI"))
    os$MSEEfficiency <- apply(os[, -4], 1, function(u) eff(u, what="OCI"))

    mse$TeXDescription <- "MSE"
    fl$TeXDescription <- paste("FLCI, $\\alpha=", fl$alpha, "$", sep="")
    os$TeXDescription <- paste("OCI,  $\\alpha=", os$alpha, "$, $\\beta=",
                               os$beta, "$", sep="")
    rbind(mse, fl, os)
}

#' Figures of optimal bias-variance trade-off and MSE efficiency
#'
#' Makes plots of optimal FLCI and OCI bias-sd ratios as function of the rate
#' exponent (one plot each), and plots of efficiency of two-sided and one-sided
#' CIs based on MSE-optimal bandwdith relative to FLCI/OCI optimal bandwidth as
#' function of the rate exponent (one plot each)
#'
#' @return Four ggplot2 figures.
#' @export
plot_BWEfficiency <- function() {
    d <- BWEfficiency(seq(0.5, 0.999, by=0.01), c(0.5, 0.8), c(0.05, 0.01))
    d2 <- BWEfficiency(seq(0.5, 0.999, by=0.01), c(0.5, 0.8),
                       c(0.1, 0.05, 0.01))
    br <- c(seq(0.5, 1, by=0.1))        # breaks in plot

    ## Fixed-length CI plots
    dfl <- d2[!grepl("OCI", d2$criterion), c("r", "t", "TeXDescription")]
    dos <- d[!grepl("FL", d$criterion), c("r", "t", "TeXDescription")]
    dfl.e <- d2[d2$criterion=="FLCI", c("r", "MSEEfficiency", "TeXDescription")]
    dos.e <- d[d$criterion=="OCI", c("r", "MSEEfficiency", "TeXDescription")]

    names(dfl) <- names(dos.e) <- names(dfl.e) <- names(dos) <-
        c("x", "y", "what")

    m1 <- list(cex=0.9, "first.qp")
    p1 <- baseplot(dfl, "", "$r$", br, br, c(0.35, 1), method=m1)
    p2 <- baseplot(dos, "", "$r$", br, br, c(0.25, 1), method=m1)
    p3 <- baseplot(dfl.e, "Relative efficiency", "$r$", br, br, c(0.3, 1),
                   method=m1)
    p4 <- baseplot(dos.e, "Relative efficiency", "$r$", br, br, c(0.5, 1.2),
                   method=list(cex=0.9, "last.qp"))

    list(p1=p1, p2=p2, p3=p3, p4=p4)
}
