tol <- .Machine$double.eps^0.75

#' All plots in this package have this form
#'
#' @param data data frame with three columns: x, y, and what
#'     (color/linetype/facet specifier)
#' @param ylab,xlab Y- and X- axis labels
#' @param facet logical
#' @param br breaks for x axis
#' @param br.lab break labels for x axis
#' @param xlim X-axis limits
#' @param method method for \code{geom_dl} direct label
#' @keywords internal
baseplot <- function(data, ylab, xlab, br=NULL, br.lab=br,
                     xlim=NULL, facet=FALSE, method=NULL) {

    p <- if (facet)
             ggplot2::qplot(x=x, y=y, geom="line", data=data,
                            color=I("#F8766D")) +
                 ggplot2::facet_grid(what ~ ., scales="free")
         else
             ggplot2::qplot(x=x, y=y, geom="line", color=what, linetype=what,
                            data=data)
    p <- p +
        ggplot2::xlab(xlab) + ggplot2::ylab(ylab) + ggplot2::theme_classic() +
        ggplot2::theme(legend.position = "none",
                       strip.background=ggplot2::element_rect(color = "white",
                           linetype=NULL),
                       axis.line.x=ggplot2::element_line(color="black",
                           size=0.2, linetype="solid"),
                       axis.line.y=ggplot2::element_line(color="black",
                           size=0.2, linetype="solid"))
    ## the axis.line.x and axis.line.y are needed as theme.classic has a bug in
    ## 2.1.0 version of ggplot2, see
    ## https://github.com/hadley/ggplot2/issues/1565

    if(!is.null(br))
        p <- p + ggplot2::scale_x_continuous(breaks=br, labels=br.lab)

    if(!is.null(xlim))
        p <- p + ggplot2::coord_cartesian(xlim = xlim)

    if (!is.null(method))
        p <-  p + directlabels::geom_dl(ggplot2::aes(label=what),
                                        data=data, method=method)
    p
}

#' Find interval containing zero of a function, then find the zero
#'
#' Given function \code{f} find \code{x0} such that \code{f(x0)==0}
#' @param f function whose root we're looking for
#' @param ival upper endpoint of initial interval in which to search
#' @param negative logical: should the lower endpoint be \code{1/ival} (if the
#'     root is guaranteed to be positive), or \code{-ival}?
#' @keywords internal
FindZero <- function(f, ival=1.1, negative=TRUE) {
    minval <- function(ival) if (negative==TRUE) -ival else min(1/ival, 1e-3)

    while(sign(f(ival))==sign(f(minval(ival))))
            ival <- 2*ival
    stats::uniroot(f, c(minval(ival), ival), tol=tol)$root
}
