context("Test optimal kernel computation in optimal[SY|Holder]Kernels.R")

itol <- .Machine$double.eps^0.75   # integration precision

test_that("Recomputing SY kernel works", {

    for (p in 1:4) {
    expect_equal(SYKernel(p, boundary=FALSE, recompute=FALSE),
                 SYKernel(p, boundary=FALSE, recompute=TRUE))
    expect_equal(SYKernel(p, boundary=TRUE, recompute=FALSE),
                 SYKernel(p, boundary=TRUE, recompute=TRUE))
    }

    ## with starting values
    expect_equal(SYKernel(3, boundary=TRUE, recompute=FALSE),
                 SYKernel(3, boundary=TRUE, recompute=TRUE,
                          start=c(-4.54028, 4.01576)))
    expect_equal(SYKernel(3, boundary=FALSE, recompute=FALSE),
                 SYKernel(3, boundary=FALSE, recompute=TRUE,
                          start=c(0, -1.92)))
    expect_equal(SYKernel(4, boundary=FALSE, recompute=FALSE),
                 SYKernel(4, boundary=FALSE, recompute=TRUE,
                          start=c(0, -2.13391, 0)))
    expect_equal(SYKernel(4, boundary=TRUE, recompute=FALSE),
                 SYKernel(4, boundary=TRUE, recompute=TRUE,
                          start=c(-7.2, 13.6, -6.8)))

})

test_that("Optimal SY kernels satisfy first-order conditions", {

    for (p in 1:4) {
        kb <- SYKernel(p=p, boundary=TRUE)$k
        ki <- SYKernel(p=p, boundary=FALSE)$k

        ## Need more subdivisions since function is zero over large range
        expect_equal(integrate(kb, 0, 1, rel.tol = itol,
                               subdivisions=1000)$value, 1)
        expect_equal(integrate(ki, 0, 1, rel.tol = itol)$value, 1/2)
        for (j in seq_len(p-1)) {
            expect_equal(integrate(function(u) u^j*kb(u), 0, 1,
                                  rel.tol = itol)$value, 0)
            expect_equal(integrate(function(u) u^j*ki(u), -1, 1,
                                   rel.tol = itol)$value, 0)
        }
    }
})

test_that("Leonov condition holds", {

    R <- function(k, boundary)
        integrate(function(u) k(u)^2, -1+boundary, 1, rel.tol=itol)$value

    leonov <- function(k, boundary, r) {
        expect_equal(R(k, boundary), k(0)*r, tolerance=1e-7)
    }

    leonov(k=HolKernel(boundary=TRUE)$k, TRUE, 4/5)
    leonov(k=HolKernel(boundary=FALSE)$k, FALSE, 4/5)

    for (p in 1:4) {
        leonov(k=SYKernel(p=p, boundary=TRUE)$k, TRUE, 2*p/(2*p+1))
        leonov(k=SYKernel(p=p, boundary=FALSE)$k, FALSE, 2*p/(2*p+1))
    }
})

test_that("Alternative computation of Holder kernel constant", {

    kb <- HolKernel(boundary=TRUE)
    ki <- HolKernel(boundary=FALSE)

    hc <- function(k) (2)^(1/5)*((4/5)*k$k(0)/k$supp)^(2/5)*(1/5)^(1/5)

    expect_equal(hc(kb), HolConstant(p=2, TRUE), tolerance=1e-7)
    expect_equal(hc(ki), HolConstant(p=2, FALSE), tolerance=1e-7)
})

test_that("Test analytic values for b in LF functions under Holder class", {

    ## Local linear, boundary
    lu <- KerEfficiency(kernel="uniform", order=1, boundary=TRUE)
    lt <- KerEfficiency(kernel="triangular", order=1, boundary=TRUE)
    le <- KerEfficiency(kernel="epanechnikov", order=1, boundary=TRUE)

    expect_equal(unname(lu["b11"]), 1/3)
    expect_equal(unname(lt["b11"]), 1/4)
    expect_equal(unname(le["b11"]), (-29+4*sqrt(106))/45, tolerance = 10e-8)

    ## Local quadratic, interior
    lu <- KerEfficiency(kernel="uniform", order=2, boundary=FALSE)
    lt <- KerEfficiency(kernel="triangular", order=2, boundary=FALSE)
    le <- KerEfficiency(kernel="epanechnikov", order=2, boundary=FALSE)

    expect_equal(unname(lu["b11"]), (-5+sqrt(105))/10)
    expect_equal(unname(lt["b11"]), (-10+sqrt(520))/30)
    root <- polyroot(c(-16, 13, 42, 21))
    expect_equal(unname(le["b11"]), Re(root[1]))

    expect_equal(unname(lu["b2"]), -1+4/sqrt(10))
    expect_equal(unname(lt["b2"]), (-2+sqrt(7))/3)
    root <- polyroot(c(5, -17, -21, -7))
    expect_equal(unname(le["b2"]), Re(root[1]))


    ## Local quadratic, boundary
    lu <- KerEfficiency(kernel="uniform", order=2, boundary=TRUE)
    lt <- KerEfficiency(kernel="triangular", order=2, boundary=TRUE)
    le <- KerEfficiency(kernel="epanechnikov", order=2, boundary=TRUE)
    expect_equal(unname(lu[c("b11", "b12")]),
                 c(4-sqrt(6), 4+sqrt(6))/10)
    expect_equal(unname(lt[c("b11", "b12")]),
                 c(5-sqrt(10), 5+sqrt(10))/15)
    root <- polyroot(c(8, -69, 54, 77))
    expect_equal(unname(le[c("b11", "b12")]), Re(root[1:2]))

})

test_that("Test analytic values for mathcal{B} under Holder class", {
    rh <- tbl_KerEfficiency("H")
    rhb <- rh$boundary
    rhi <- rh$interior

    ints <- function(s) integrate(s, 0, 1, rel.tol=tol)$value


    expect_equal(rhb$B1[rhb$order==0], c(1/2, 1/3, 3/8))

    se <- function(b) abs((b-1)^2*(-45*b^2-58*b+19)/19)
    expect_equal(rhb$B1[rhb$order==1], c(8/27, 27/128, ints(se)))
    ## function s_{0}(k^*_2)
    su <- function(b) abs((b-1)*(-10*b^2+8*b-1))
    st <- function(b) abs((b-1)^2*(15*b^2-10*b+1))
    se <- function(b) abs((b-1)^2/8*(77*b^3+54*b^2-69*b+8))
    expect_equal(rhb$B1[rhb$order==2], c(ints(su), ints(st), ints(se)))

    expect_equal(rhb$B2[rhb$order==1], c(1/6, 1/10, 11/95))
    expect_equal(rhb$B2[rhb$order==2], c(216/3125, 32/729,
    (50903461212000-308738237600*sqrt(27105))/1458956660623))

    expect_equal(rhb$B3[rhb$order==2], c(1/20, 1/35, 15/448))

    expect_equal(rhi$B1[rhb$order==0], c(1/2, 1/3, 3/8))
    expect_equal(rhi$B1[rhb$order==1], c(1/2, 1/3, 3/8))
    su <- function(b) abs((b-1)*(5*b^2+5*b-4)/4)
    st <- function(b) abs((b-1)^2*(-15*b^2-10*b+7)/7)
    se <- function(b) abs((b-1)^2*(-21*b^3-42*b^2-13*b+16)/16)
    expect_equal(rhi$B1[rhb$order==2], c(ints(su), ints(st), ints(se)))

    expect_equal(rhi$B2[rhb$order==1], c(1/3, 1/6, 1/5))
    su <- function(b) 2*abs((b-1)^2*(-5*b^2-10*b+3)/16)
    st <- function(b) 2*abs((b-1)^3*(3*b^2+4*b-1)/7)
    se <- function(b) 2*abs((b-1)^3*(7*b^3+21*b^2+17*b-5)/32)
    expect_equal(rhi$B2[rhb$order==2], c(ints(su), ints(st), ints(se)))

    expect_equal(rhi$B3[rhb$order==2], c(1/16, 8/245, 5/128))



})
