## This file replicates all calculations in the empirical illustration in
## Section 4 and Appendices C.1 and E.3 in the supplement.

## Brief description of output:
##
## - lee-0.00002-0.1-117-IKEHW-nn.RData: results for confidence interval
##   calculations on a grid of C-values
## - lee-results-...TIMESTAMP: log file of results of calculations referred to
##   in text.
## - lee-0.0002-0.1-1007-IKEHW-nn-ll...tex graphs for estimation with local
##   linear regression and triangular kernel
## - lee-0.0002-0.1-1007-IKEHW-nn-opt...tex graph for estimation with optimal
##   SY kernel


################## PARAMETERS ##################

## Use these parameters for Section 4.2
se.method <- "supplied.var"     # Use nearest neighbor to estimate variance
fig.prefix <- NULL

## Use these parameters for all results except Section 4.2
## se.method <- "nn"     # Use nearest neighbor to estimate variance
## fig.prefix <- "results/lee"

job <- 4              # Specify how fine grid of smoothness parameters should be
se.initial <- "IKEHW" # Use IK bw + EHW variance for initial variance estimate

################################

if (job==2) {
    ## Takes about 30 minutes
    Cs <- c(seq(0.00002, 0.0001, by=0.00001),
            seq(0.0002, 0.001, by=0.0001),
            seq(0.002, 0.1, by=0.001))
    ## indices of ticks for graphs
    C.ticks <- c(1, 37, 57, 77, 97, 117)
} else if (job==4) {
    ## Takes about 8 hours
    Cs <- c(seq(2, 4, by=0.22), seq(4, 1000, by=1))/100^2
    C.ticks <- c(1, 207, 407, 607, 807, 1007) # indices of ticks for graphs
}

d.lee <- RDHonest::RDData(RDHonest::lee08, cutoff=0)
filen <- paste0(min(Cs), "-", max(Cs), "-",
                length(Cs), "-", se.initial, "-", se.method)

cat("Started at: ", format(Sys.time(), "%Y-%m-%d-%H.%M"), "\n")
sink(paste0("results/lee-results-", filen, "-",
            format(Sys.time(), "%Y-%m-%d-%H.%M")), split=TRUE)

r <- NPRHonest::FullAnalysis(d.lee, Cs, save.prefix="results/lee",
                             fig.prefix=fig.prefix, se.initial=se.initial,
                             ylab="Electoral Advantage (\\%)", C.ticks=C.ticks,
                             kern="triangular", se.method=se.method)

## Numbers used in text
if (se.method=="nn"){
    cat("==== Section 4.1 ====\n")
    cat("Range of global minimax MSE estimates for C>0.005: ",
        range(r$dopt$e$estimate[r$dopt$e$C>0.005]), "\n")


    ## Now C-bound
    d <- RDHonest::RDPrelimVar(d.lee, se.initial="NN")
    r <- RDHonest::RDSmoothnessBound(d, s=200, separate=TRUE,
                                 multiple=FALSE, sclass="T")
    ## print(c(r$ne$hatM/2, r$ne$lower/2))
    ## print(c(r$po$hatM/2, r$po$lower/2))
    cat("\n==== Appendix E.3 ====\n",
        "Divide the estimates by 2 to express in terms of C=2*M\n")
    print(r)
} else if (se.method=="supplied.var") {
    cat("==== Section 4.2 ====\n")
    cat("Bounds on adaptation: Range for OCI8, OCI5, FLCI, FLCI[C>0.0001]: \n",
        range(r$dopt$eoci8[[1]]), "\n",
        range(r$dopt$eoci5[[1]]), "\n",
        range(r$dopt$eflci[[1]]), "\n",
        range(r$dopt$eflci[r$dopt$eflci$C>=0.0001, 1]), "\n")
    cat("Optimal delta range for OCI and FLCI:\n",
        range(r$dopt$flci$delta), "\n",
        range(r$dopt$oci$delta), "\n")
    cat("Range of eff.obs, hm, hm ratios for OCI / FLCI: \n",
        range(r$dopt$oci$eff.obs/r$dopt$flci$eff.obs), "\n",
        range(r$dopt$oci$hm/r$dopt$flci$hm), "\n",
        range(r$dopt$oci$hp/r$dopt$flci$hp), "\n")
    cat("Range of eff.obs, hm, hm ratios for FLCI / MSE: \n",
        range(r$dopt$flci$eff.obs/r$dopt$e$eff.obs), "\n",
        range(r$dopt$flci$hm/r$dopt$e$hm), "\n",
        range(r$dopt$flci$hp/r$dopt$e$hp), "\n")

    ELratio <- r$dopt$oci$omega /
        (r$dopt$e$sd*(qnorm(0.8)+qnorm(0.95)-r$dopt$e$delta)+r$dopt$e$omega)
    cat("Minimum efficiency of MSE-based FLCI, OCI: ",
        100*c(min(r$dopt$flci$hl/r$dopt$e$hl), min(ELratio)), "\n")

    ELratio <- (2*r$dopt$oci$maxbias+r$dopt$oci$sd*(qnorm(0.95)+qnorm(0.8))) /
        (2*r$d1$oci$maxbias+r$d1$oci$sd*(qnorm(0.95)+qnorm(0.8)))
    cat("Minimum efficiency of LL-based FLCI, OCI:",
        100*c(min(r$dopt$flci$hl/r$d1$flci$hl), min(ELratio)), "\n")
}

sink()

cat("Finished at: ", format(Sys.time(), "%Y-%m-%d-%H.%M"), "\n")
